////////////////////////////////////////////////////////////////////// 
// mach2dat/Main.cpp 
// (c) 2008 Yun Li
// 
// March 15, 2008
//

#include "MathCholesky.h"
#include "LogisticRegression.h"
#include "MathStats.h"
#include "MathSVD.h"
#include "InputFile.h"
#include "IntArray.h"
#include "Error.h"
#include "StringHash.h"
#include "Parameters.h"
#include "Pedigree.h"
#include "MemoryAllocators.h"

#include "simplex.h"
#include <ctime>

double llk(double * beta, int p);

double ** xpllk;
double *yllk;
int rowdim;
int coldim;

int main(int argc, char ** argv)
{
setbuf(stdout, NULL);
setbuf(stderr, NULL);
  clock_t startt, endt;
  startt = clock();

	// ("D:/tmp/codes/cpp/mach2dat/examples/collinearity/pheno.noGender.dat") ;
	// ("D:/tmp/codes/cpp/mach2dat/examples/collinearity/pheno.withGender.dat") ;
	// ("D:/tmp/codes/cpp/mach2dat/examples/collinearity/pheno.allMales.ped") ;
	
	// ("D:/tmp/codes/cpp/mach2dat/examples/collinearity/rs4784323.mlinfo") ;
	// ("D:/tmp/codes/cpp/mach2dat/examples/collinearity/rs4784323.mldose") ;
	String datfile ("pheno.dat");
	String pedfile ("pheno.ped");
	String infofile ("sample.mlinfo");
	String dosefile ("sample.mldose");
	String outfile;
	bool covariates = true, lr = true;
	bool samplesize = true, freq = false;
	bool verboseSampleSize = false;
	String method = "newton"; // newton or simplex
        double rsqcutoff = 0.0001;
	int nrrounds = 20;
	
#ifndef VERSION
   printf("mach2dat 1.0 -- Disease-snp Association Tests with Imputed Dosages\n"
#else
   printf("mach2dat " VERSION " -- Disease-snp Association Tests with Imputed Dosages\n"
#endif
          "(c) 2008 Yun Li, Wei Chen, Goncalo Abecasis \n\n");

	ParameterList pl;
	
BEGIN_LONG_PARAMETERS(longParameters)
   LONG_PARAMETER_GROUP("Phenotypic Data")
      LONG_STRINGPARAMETER("datfile", &datfile)
      LONG_STRINGPARAMETER("pedfile", &pedfile)
   LONG_PARAMETER_GROUP("Imputed Genotype Data")
      LONG_STRINGPARAMETER("infofile", &infofile)
      LONG_STRINGPARAMETER("dosefile", &dosefile)
   LONG_PARAMETER_GROUP("Analysis Options")
      LONG_PARAMETER("useCovariates", &covariates)
      LONG_PARAMETER("likelihoodratio", &lr)
      LONG_PARAMETER("samplesize", &samplesize)
      LONG_PARAMETER("verboseSampleSize", &verboseSampleSize)
      LONG_INTPARAMETER("nrrounds", &nrrounds)
      LONG_DOUBLEPARAMETER("rsqcutoff", &rsqcutoff)
      LONG_STRINGPARAMETER("method",&method)  // wc
   LONG_PARAMETER_GROUP("Output")
      LONG_PARAMETER("frequency", &freq)
END_LONG_PARAMETERS();

	pl.Add(new LongParameters("Available Options", longParameters));
  
	pl.Add(new HiddenString('d', "datfile", datfile));
	pl.Add(new HiddenString('p', "pedfile", pedfile));
	pl.Add(new HiddenString('i', "infofile", infofile));
	pl.Add(new HiddenString('g', "dosefile", dosefile));
	pl.Add(new HiddenString('g', "method", method)); // wc
	//pl.Add(new HiddenString('o', "outfile", outfile));
  
	pl.Read(argc, argv);
	pl.Status();

	if (verboseSampleSize) samplesize = false;
  
	// Load pedigree data
	Pedigree ped;
  
	ped.Prepare(datfile);
	ped.Load(pedfile);

	// Load marker information
	printf("\nLoading marker information ...\n");
  
	StringArray markerlist;
	markerlist.Read(infofile);
	for (int i = markerlist.Length() - 1; i >= 0; i--)
	{
		markerlist[i].Trim();
		if (markerlist[i].IsEmpty())
			markerlist.Delete(i);
	}
  
	printf("   %d markers will be analyzed\n\n", markerlist.Length() - 1);
  
	// Process dosage file
	printf("Processing dosage file ...\n\n");
  
	IFILE f = ifopen(dosefile, "rb");

	if (f == NULL)
		error("Dosage file %s can not be opened\n\n", (const char *) dosefile);

	String      buffer, famid, pid;
	StringArray tokens;

	// Go through dosage file to mark individuals with also found in pedigree;
	StringHash    imputed;
  
	while (!ifeof(f))
	{
		tokens.ReplaceTokens(buffer.ReadLine(f));
    
		if (tokens.Length() == 0) continue;
    
		if (tokens.Length() != markerlist.Length() + 1)
		{
			printf("Expecting %d markers, but found %d markers for individuals %s\n",
				markerlist.Length() - 1, tokens.Length() - 2, (const char *) tokens[0]);
			continue;
		}
    
		int arrow = tokens[0].Find("->");
    
		if (arrow < 0)
		{
			//printf("Labels should be FAMILY->INDIVIDUAL, cannot decode individual %s\n", (const char *) tokens[0]);
			continue;
		}
    
		famid = tokens[0].Left(arrow);
		pid = tokens[0].SubStr(arrow + 2);
    
		Person * person = ped.FindPerson(famid, pid);
    
		if (person == NULL)
		{
			// printf("Individual %s is not in pedigree\n", (const char *) tokens[0]);
			continue;
		}

		imputed.Add(tokens[0]);
	}


	if (imputed.Entries() == 0)
	{
		printf("No individuals with both dosage and phenotpe information \n\n");
		return 0;
	}

	printf("\n%d individuals are found in both dosage and phenotype files\n\n", imputed.Entries());
  
	if (!covariates)
	{
		ped.covariateCount = 0;
	}
  
	LogisticRegression lrengine; // our logistic regression engine

	// If there are no affection status(es), nothing to do
	if (ped.affectionCount == 0)
		return 0;

	// Check for related individuals in the sample
	// 
	for (int i = 0; i < ped.count; i++)
		if (!ped[i].isFounder() && false)
			error ("Your sample includes related individuals\n\n");

	// Array of informative individuals;
	IntArray      pheno; 
	StringIntHash peopleindexhash; // key: famid->pid & value: index in ped;

	// This is the phenotypic vector for each affection status
	Vector y;

	// This is the full design matrix for each affection status;
	// # parameters: slope + # covariates + SNPofInterest;
	Matrix X;

	// This is the reduced design matrix for each affection status;
	// # parameters: slope + # covariates ;
	Matrix Xnull;

	printf("\n");
	        printf("        \tINFORMATION FROM .info FILE              Disease ASSOCIATION\n");
	if (lr)
	{
		if (verboseSampleSize)
		{
			printf("        \t======================================   =============================================================================\n");
			printf("TRAIT   \tMARKER          ALLELES  FREQ1    RSQR   EFFECT1  OR      STDERR  WALDCHISQ PVALUE     LRCHISQ LRPVAL NCASES NCONTROLS\n");
		}
		else
		{
			printf("        \t======================================   ============================================================\n");
			printf("TRAIT   \tMARKER          ALLELES  FREQ1    RSQR   EFFECT1  OR      STDERR  WALDCHISQ PVALUE     LRCHISQ LRPVAL\n");
		}
	}
	else
	{
		if (verboseSampleSize)
		{
			printf("        \t======================================   ==========================================================\n");
			printf("TRAIT   \tMARKER          ALLELES  FREQ1    RSQR   EFFECT1  OR      STDERR  WALDCHISQ PVALUE NCASES NCONTROLS\n");
		}
		else
		{
			printf("        \t======================================   =========================================\n");
			printf("TRAIT   \tMARKER          ALLELES  FREQ1    RSQR   EFFECT1  OR      STDERR  WALDCHISQ PVALUE\n");
		}
	}
	

	 /////////////////////  Newton's Method /////////////////////////////// 
   if (method == "newton")
   {
	// Loop through affection statuses in the pedigree
	for (int a = 0; a < ped.affectionCount; a++)
	{
		// Start with empty list of phenotyped individuals
		pheno.Clear();
		peopleindexhash.Clear();

		// Then, list individuals in the sample with affection status information and  
		// all necessary covariates

		int indexinteger = 0; // fix peopleindexhash bug 

		for (int i = 0; i < ped.count; i++)
		{
			if (ped[i].isDiagnosed(a) && (!covariates || ped[i].isFullyControlled()) 
				&& (imputed.Find(ped[i].famid+"->"+ped[i].pid) != -1))
			{
				pheno.Push(i);
				//printf("i = %d index = %d id = %s int = %d\n",
					//i,indexinteger,(const char *)ped[i].pid, pheno[pheno.Last()]);

				// fix peopleindexhash bug 
				//peopleindexhash.Add(ped[i].famid+"->"+ped[i].pid,pheno[pheno.Last()]);
				peopleindexhash.Add(ped[i].famid+"->"+ped[i].pid,indexinteger++);
			}
		} // end of individual i


		if (pheno.Length() == 0)
		{
			printf("Affection Status: %s -- No individuals with phenotype and covariate data\n\n",
				(const char *) ped.affectionNames[a]);
			continue;
		}

		if (covariates && ped.covariateCount && pheno.Length() <= ped.covariateCount + 2)
		{
			error("There are %d covariates in the model, but only %d complete observations for affection status %s\n",
				ped.covariateCount, pheno.Length(), (const char *) ped.affectionNames[a]);
		}
        
     	// Allocate phenotype vector
		y.Dimension(pheno.Length());

		// Allocate design matrices
		X.Dimension(pheno.Length(), 2 + (covariates ? ped.covariateCount : 0));
		Xnull.Dimension(pheno.Length(), 1 + (covariates ? ped.covariateCount : 0));

                int ncases = 0;
                int ncontrols = 0;

		// Fill in design matrix (with covariate values), and y vector;
		for (int i = 0; i < pheno.Length(); i++)
		{
			y[i] = ped[pheno[i]].affections[a] - 1;
                        if (y[i]) ncases ++;
                        else ncontrols ++;

			// The first column of the design matrix is constant one, for the slope
			X[i][0] = Xnull[i][0] = 1.0;

			// The next columns of the design matrix refer to individual covariates
			if (covariates)
				for (int j = 1; j <= ped.covariateCount; j++)
				{
					X[i][j] = Xnull[i][j] = ped[pheno[i]].covariates[j-1];
				}
		} // fill individual i to y and design matrix

                if (samplesize) printf("%-8s\tSAMPLE-SIZE\t%d cases\t%d controls\n", (const char *) ped.affectionNames[a], ncases, ncontrols);

		// if likelihood ratio test is desired, fit a NULL model with covariates only;
		double nulldeviance = 0.0;
		bool nullmodel = false;
		if (lr)
		{
			nullmodel = lrengine.FitLogisticModel(Xnull, y, nrrounds);
			if (nullmodel)
				{nulldeviance = lrengine.GetDeviance(Xnull, y);}
			//else
				//{printf("Failed to fit null model\n"); }
		}

		// Fill in the design matrix with dosages for SNP of interest;
		// rewind dosage file for each trait (tmb?)
		ifrewind(f);
		// Matrix Dosages;
		// Dosages.Dimension(pheno.Length(), (markerlist.Length() - 1));
		// Dosages.Zero();
		float ** Dosages = AllocateFloatMatrix(pheno.Length(), (markerlist.Length() - 1));

		while (!ifeof(f))
		{
			tokens.ReplaceTokens(buffer.ReadLine(f));

			if (tokens.Length() == 0) continue;

//printf("%s\n",(const char *) tokens[0]);

			if (tokens.Length() != markerlist.Length() + 1)
			{
				printf("Expecting %d markers, but found %d markers for individual %s\n",
					markerlist.Length() - 1, tokens.Length() - 2, (const char *) tokens[0]);
				continue;
			}

			int index = peopleindexhash.Find(tokens[0]);

			if (index == -1)
				continue;

			index = peopleindexhash.Integer(index);

			if (index < 0)
				continue;

			for (int j = 0; j < (markerlist.Length() - 1); j++)
			{
				Dosages[index][j] = tokens[j+2].AsDouble();
			}
		}// end of while reading dosage file

		for (int j = 0; j < (markerlist.Length() - 1); j++)
		{
  		int ncasesSNP = ncases;
  		int ncontrolsSNP = ncontrols;
  		int numMissingDosages = 0;
  		Vector ySNP;
  		Matrix XSNP;
  		Matrix XnullSNP;
		double casedosages = 0.0, controldosages = 0.0;
		  //ySNP.Dimension(pheno.Length());
  		//XSNP.Dimension(pheno.Length(), 2 + (covariates ? ped.covariateCount : 0));
		  //XnullSNP.Dimension(pheno.Length(), 1 + (covariates ? ped.covariateCount : 0));
			for (int k = 0; k < pheno.Length(); k++)
			{
				X[k][ped.covariateCount + 1] = Dosages[k][j];
				if (Dosages[k][j] == -9)
  				{numMissingDosages ++;continue;}
				if (freq)
				{
					if (y[k]) casedosages += Dosages[k][j];
					else controldosages += Dosages[k][j];
				}
			}
			if (numMissingDosages > 0)
			{
  			numMissingDosages = 0;
    		XSNP = Matrix(X);
    		XnullSNP = Matrix(Xnull);
    		ySNP = Vector(y);
  			for (int k = 0; k < pheno.Length(); k++)
  			{
  				if (Dosages[k][j] == -9)
  				{
    				int newindex = k - numMissingDosages;
    				if (y[k]) ncasesSNP --;
    				else ncontrolsSNP --;
    				XSNP.DeleteRow(newindex);
    				XnullSNP.DeleteRow(newindex);
    				ySNP.DeleteDimension(newindex);
    				numMissingDosages ++;
  				}
  			}
			}
			double nulldevianceSNP;
			if (numMissingDosages > 0)
			{
  			// need to re-fit null model
  			if (lr)
  			{
    			bool nullmodelSNP = false;
    			nullmodelSNP = lrengine.FitLogisticModel(XnullSNP, ySNP, nrrounds);
    			if (nullmodelSNP)
    				{nulldevianceSNP = lrengine.GetDeviance(XnullSNP, ySNP);}
  			}
			} // re-fit null model
			else
			{
  			nulldevianceSNP = nulldeviance;
			}

			double casefreq = casedosages * 0.5 / ncasesSNP;
			double controlfreq = controldosages * 0.5 / ncontrolsSNP;

			tokens.ReplaceTokens(markerlist[j+1]);
			printf("%-8s\t%-16s", (const char *) ped.affectionNames[a], (const char *) tokens[0]);

			if (tokens.Length() < 5)
				printf("** Marker Info Error **  ");
			else
				printf("%-9s", (const char *) (tokens[1] + "," + tokens[2]));

			if(tokens[3].AsDouble() >= 1.0)
				printf("1.000");
			else
				printf(".%04.0f", tokens[3].AsDouble() * 10000.0);

			if (freq)
			{
				printf("(.%04.0f,.%04.0f)", casefreq * 10000.0, controlfreq * 10000.0);
			}

			if(tokens[6].AsDouble() >= 1.0)
				printf("   1.000   ");
			else
				printf("   .%04.0f   ", tokens[6].AsDouble() * 10000.0);

			bool fullmodel = false;
			if ( numMissingDosages > 0) fullmodel = lrengine.FitLogisticModel(XSNP, ySNP, nrrounds);
			else                        fullmodel = lrengine.FitLogisticModel(X, y, nrrounds);

			if (fullmodel)
			{
				// perform Wald test
				int SNPindex = ped.covariateCount + 1;
				double effect = lrengine.B[SNPindex] ;
				double variance = lrengine.covB[SNPindex][SNPindex] + 1e-30;
				double serror = sqrt(variance);
				double waldchisq = effect / serror;
				waldchisq *= waldchisq;

				int digits = (effect < 999.9995 && effect > -99.9995) ? 3 :
							 (effect < 9999.995 && effect > -999.995) ? 2 :
							 (effect < 99999.95 && effect > -9999.95) ? 1 : 0;

				int sdigits = (serror < 999.9995 && serror > -99.9995) ? 3 :
							  (serror < 9999.995 && serror > -999.995) ? 2 :
							  (serror < 99999.95 && serror > -9999.95) ? 1 : 0;

				// when Rsq small
				if (tokens[6].AsDouble() > rsqcutoff)
				//if (tokens[6].AsDouble() != 0)
					printf("% -8.*f %-7.*f %-7.*f %-9.4f %-11.4g",
						digits, effect, digits, exp(effect), sdigits, serror,
						waldchisq, waldchisq > 0 ? chidist(waldchisq,1) : 1.0);
				else
					printf(" NA      NA      NA      NA        NA  ");


				if (lr)
				{
					if (tokens[6].AsDouble() > rsqcutoff)
					//if (tokens[6].AsDouble() != 0)
					{
						double fulldeviance;
						if (numMissingDosages > 0) fulldeviance = lrengine.GetDeviance(XSNP, ySNP);
						else                       fulldeviance = lrengine.GetDeviance(X, y);
						
						double lrchisq = nulldevianceSNP - fulldeviance;

						printf("%-7.4f %-7.4g",
								lrchisq, lrchisq > 0 ? chidist(lrchisq,1) : 1.0);
					}
					else
						printf("        NA      NA");
				}
			} // end if fullmodel

			else // fullmodel not fitted
			{
				printf(" NA      NA      NA      NA        NA  ");
				if (lr)
					printf("        NA      NA");
			}

			if (verboseSampleSize)
				printf("%6d %9d", ncasesSNP, ncontrolsSNP);

			printf("\n");

		} // for marker j;

		if (Dosages != NULL) FreeFloatMatrix(Dosages, pheno.Length());
	
	} // end of affection status a

} // end of if (method == "newton")

///////////////////// Simplex Method /////////////////////////////

  if (method == "simplex")
  {
   // Loop through affection statuses in the pedigree
	for (int a = 0; a < ped.affectionCount; a++)
	{
		// Start with empty list of phenotyped individuals
		pheno.Clear();

		// Then, list individuals in the sample with affection status information and  
		// all necessary covariates

		int indexinteger = 0; // fix peopleindexhash bug 

		for (int i = 0; i < ped.count; i++)
		{
			if (ped[i].isDiagnosed(a) && (!covariates || ped[i].isFullyControlled()) 
				&& (imputed.Find(ped[i].famid+"->"+ped[i].pid) != -1))
			{
				pheno.Push(i);
				//printf("i = %d index = %d id = %s int = %d\n",
					//i,indexinteger,(const char *)ped[i].pid, pheno[pheno.Last()]);

				// fix peopleindexhash bug 
				//peopleindexhash.Add(ped[i].famid+"->"+ped[i].pid,pheno[pheno.Last()]);
				peopleindexhash.Add(ped[i].famid+"->"+ped[i].pid,indexinteger++);
			}
		} // end of individual i


		if (pheno.Length() == 0)
		{
			printf("Affection Status: %s -- No individuals with phenotype and covariate data\n\n",
				(const char *) ped.affectionNames[a]);
			continue;
		}

		if (covariates && ped.covariateCount && pheno.Length() <= ped.covariateCount + 2)
		{
			error("There are %d covariates in the model, but only %d complete observations for affection status %s\n",
				ped.covariateCount, pheno.Length(), (const char *) ped.affectionNames[a]);
		}
        
        
		// if likelihood ratio test is desired, fit a NULL model with covariates only;
		double nulldeviance = 0.0;
		double fulldeviance = 0.0;
		bool nullmodel = false;
		
		// Fill in the design matrix with dosages for SNP of interest;
		// rewind dosage file for each trait (tmb?)
		ifrewind(f);
		// Matrix Dosages;
		// Dosages.Dimension(pheno.Length(), (markerlist.Length() - 1));
		// Dosages.Zero();
		float ** Dosages = AllocateFloatMatrix(pheno.Length(), (markerlist.Length() - 1));

		while (!ifeof(f))
		{
			tokens.ReplaceTokens(buffer.ReadLine(f));

			if (tokens.Length() == 0) continue;

//printf("%s\n",(const char *) tokens[0]);

			if (tokens.Length() != markerlist.Length() + 1)
			{
				printf("Expecting %d markers, but found %d markers for individual %s\n",
					markerlist.Length() - 1, tokens.Length() - 2, (const char *) tokens[0]);
				continue;
			}

			int index = peopleindexhash.Find(tokens[0]);

			if (index == -1)
				continue;

			index = peopleindexhash.Integer(index);

			if (index < 0)
				continue;

			for (int j = 0; j < (markerlist.Length() - 1); j++)
			{
				Dosages[index][j] = tokens[j+2].AsDouble();
			}
		}// end of while reading dosage file


                int ncases = 0;
                int ncontrols = 0;

        Simplex simplex(pheno.Length(),ped.covariateCount+2);
       // Fill in design matrix (with covariate values), and y vector;
	   for (int i = 0; i < pheno.Length(); i++)
		{
			//simplex.y[i] = ped[pheno[i]].affections[a] - 1;
			simplex.y[i] = ped[pheno[i]].affections[a]-1;
			if (simplex.y[i]) ncases ++;
			else ncontrols ++;

			// The first column of the design matrix is constant one, for the slope
			simplex.xp[i][0]  = 1.0;

			// The next columns of the design matrix refer to individual covariates
		} // fill individual i to y and design matrix

                if (samplesize) printf("%-8s\tSAMPLE-SIZE\t%d cases\t%d controls\n", (const char *) ped.affectionNames[a], ncases, ncontrols);       

	   double * beta   = new double [ped.covariateCount+2];
	
		for (int j = 0; j < (markerlist.Length() - 1); j++)
	    { if (covariates)
		      {    for (int i = 0; i < pheno.Length(); i++)
		    	     for (int k = 1; k <= ped.covariateCount; k++)
					  simplex.xp[i][k] = ped[pheno[i]].covariates[k-1];
			  }
		 		
		  for (int i = 0; i < pheno.Length(); i++)
		     simplex.xp[i][ped.covariateCount+1] = Dosages[i][j];
		
		   rowdim = pheno.Length(); coldim = ped.covariateCount + 2;
           for (int t = 0;t < coldim;t++) beta[t]=0;
		   
		   simplex.count = 0;
		   xpllk = simplex.xp;
		   yllk = simplex.y;
	       fulldeviance = 2 * simplex.amoeba(beta,coldim,pheno.Length(),llk,1e-15);


         
		    tokens.ReplaceTokens(markerlist[j+1]);
			printf("%-8s\t%-15s", (const char *) ped.affectionNames[a], (const char *) tokens[0]);

			if (tokens.Length() < 5)
				printf("** Marker Info Error **  ");
			else
				printf("%-9s", (const char *) (tokens[1] + "," + tokens[2]));

			if(tokens[3].AsDouble() >= 1.0)
				printf("1.000   ");
			else
				printf(".%04.0f   ", tokens[3].AsDouble() * 10000.0);

			if(tokens[6].AsDouble() >= 1.0)
				printf("1.000   ");
			else
				printf(".%04.0f   ", tokens[6].AsDouble() * 10000.0);

			//printf ("\n");

    
	// obtain covariance matrix to perform Wald test
	// covB = solve(t(X)%*%V%*%X)
	
	// Transpose X and multiply by diagonal V
	X.Dimension(pheno.Length(), 2 + (covariates ? ped.covariateCount : 0));
	X.Zero();
    
	for (int i = 0; i < rowdim; i ++)
		for (int j = 0; j < coldim; j++)
			X[i][j] = simplex.xp[i][j];

	Matrix covB;
	covB.Dimension(coldim, coldim);
	covB.Zero();

    Vector p, V;
	p.Dimension(rowdim);
	V.Dimension(rowdim);
	p.Zero();
	V.Zero();

    for (int i = 0; i < rowdim; i++)
		{
			double d = 0;
			for (int k = 0; k < coldim; k ++)
			{
				d += beta[k] * X[i][k];
			} // for parameter beta-k

			p[i] = 1.0 / (1.0 + exp(-d));
			V[i] = p[i] * (1 - p[i]);
		} // for observation i
       
	Matrix XtV;
	XtV.Dimension(coldim, rowdim);
	XtV.Zero();
	for (int i = 0; i < rowdim; i++)
		for (int k = 0; k < coldim; k++)
			XtV[k][i] = X[i][k] * V[i];

	Matrix XtVX;
	XtVX.Dimension(X.cols, X.cols);
	XtVX.Zero();
	XtVX.Product(XtV,X);

	covB.Zero();
	covB = XtVX;
	//SVD svd;
	//svd.SVDInverse(covB);
	Cholesky chol;
	if (!chol.TryDecompose(covB))
		return false;
	chol.Decompose(covB);
	chol.Invert();
	covB = chol.inv;



          //if (fullmodel)
	     if (1)
			{
				// perform Wald test
				double effect = beta[ped.covariateCount + 1] ;
				double variance = covB[ped.covariateCount + 1][ped.covariateCount + 1] + 1e-30;
				double serror = sqrt(variance);
				double waldchisq = effect / serror;
				waldchisq *= waldchisq;

				int digits = (effect < 999.9995 && effect > -99.9995) ? 3 :
							 (effect < 9999.995 && effect > -999.995) ? 2 :
							 (effect < 99999.95 && effect > -9999.95) ? 1 : 0;

				int sdigits = (serror < 999.9995 && serror > -99.9995) ? 3 :
							  (serror < 9999.995 && serror > -999.995) ? 2 :
							  (serror < 99999.95 && serror > -9999.95) ? 1 : 0;


				// when Rsq small
				if (tokens[6].AsDouble() > rsqcutoff)
				//if (tokens[6].AsDouble() != 0)
					printf("% -8.*f %-7.*f %-9.4f %-11.4g",
						digits, effect, sdigits, serror,
						waldchisq, waldchisq > 0 ? chidist(waldchisq,1) : 1.0);
				else
					printf(" NA      NA      NA        NA  ");


				if (lr)
				{   Simplex simplex0(pheno.Length(),ped.covariateCount+1);
	                double * beta0 = new double [ped.covariateCount+1];
					for (int i = 0; i < pheno.Length(); i++)
	                  {	simplex0.y[i] = ped[pheno[i]].affections[a]-1;

			            simplex0.xp[i][0]  = 1.0;
              	} 					
					 if (covariates)
		              {    for (int i = 0; i < pheno.Length(); i++)
		    	              for (int k = 1; k <= ped.covariateCount; k++)
					            simplex0.xp[i][k] = ped[pheno[i]].covariates[k-1];
			          }
		              rowdim = pheno.Length(); coldim = ped.covariateCount + 1;
                      for (int t = 0;t < coldim;t++) beta0[t]=0;
		   
		              simplex0.count = 0;
		              xpllk = simplex0.xp;
		              yllk = simplex0.y;
	                  double nulldeviance = 2 * simplex0.amoeba(beta0,coldim,pheno.Length(),llk,1e-15);
                   
					if (tokens[6].AsDouble() > rsqcutoff)
					//if (tokens[6].AsDouble() != 0)
					{   
						//printf("null = %lf full = %lf ",nulldeviance, fulldeviance);
						double lrchisq = nulldeviance - fulldeviance;

						printf("%-7.4f %-7.4g",
								lrchisq, lrchisq > 0 ? chidist(lrchisq,1) : 1.0);
					}
					else
						printf("        NA      NA");
				}
			} // end if fullmodel
        
			else // fullmodel not fitted
			{
				printf(" NA      NA      NA        NA");
				if (lr)
					printf("         NA      NA");
			}


                        if (verboseSampleSize)
                                printf("%6d %9d", ncases, ncontrols);
      
			printf("\n");



		} // for marker j;

		   
	
///////////////////// end of simplex //////////////////////////////

		if (Dosages != NULL) FreeFloatMatrix(Dosages, pheno.Length());


	} // end of affection status a

 }// end of if (method == "simplex")

  endt = clock();
  int lapsetime = (int) ((double)(endt - startt) / CLOCKS_PER_SEC);
  printf("\nAnalysis took %d seconds\n\n", lapsetime);


	ifclose(f);
}


double llk(double * beta, int p){ //p is the dimension of covariates including 1
	int i,j;
	double f1,f2=0.0;
	for(i=0;i<rowdim;i++){
		f1=0.0;
		for(j=0;j<coldim;j++)
			f1+=beta[j]* xpllk[i][j];
		f2+=yllk[i]*f1-log(1+exp(f1));
	}
	return -f2;
}
