#include "simplex.h"
#include<stdio.h>
#include<math.h>
#include<stdlib.h>


#define ZEPS 1e-10

Simplex::Simplex(int rowdim, int coldim)
{ count = 0;
  y = alloc_vector(rowdim);
  xp = alloc_matrix(rowdim,coldim);
}

Simplex::~Simplex()
{

}

// allocating vectors
double * Simplex::alloc_vector(int cols){  
	return (double *) malloc(sizeof(double) * cols);
}


void Simplex::free_vector(double * vector, int cols){
	free(vector);
}

// allocating matrices
double ** Simplex::alloc_matrix(int rows, int cols){
	int i;
	double ** matrix = (double **) malloc(sizeof(double *) * rows);

	for (i=0;i<rows;i++)
		matrix[i]=alloc_vector(cols);
	return matrix;
}

void Simplex::free_matrix(double ** matrix, int rows, int cols){
	int i;
	for(i=0;i<rows;i++)
		free_vector(matrix[i], cols);
	free(matrix);
}


// creating a simplex
double ** Simplex::make_simplex(double * point, int dim){
	int i,j;
	double ** simplex = alloc_matrix(dim+1,dim);

	for (i=0; i<dim+1; i++)
		for (j=0; j<dim; j++)
			simplex[i][j] = point[j];

	for (i=0; i<dim; i++)
		simplex[i][i]+=1.0;

	return simplex;
}

// evaluating functions at vertices
void Simplex::evaluate_simplex(double ** simplex, int dim, double * fx, double (*func)(double *, int)){
	int i;
	for (i=0; i<dim+1; i++)
		fx[i]=(*func)(simplex[i],dim);
}

// finding extremes
void Simplex::simplex_extremes(double *fx, int dim, int & ihi, int & ilo, int & inhi){
	int i;

	if (fx[0]>fx[1])
	   { ihi=0; ilo=inhi=1; }
	else
	   { ihi=1; ilo=inhi=0; }
	
	for (i=2; i<dim+1; i++)
		if (fx[i]<=fx[ilo])
			ilo=i;
		else if (fx[i]>fx[ihi])
		{ inhi = ihi; ihi=i; }
		else if (fx[i]>fx[inhi])
			inhi=i;
}

// direction for optimization
void Simplex::simplex_bearings (double ** simplex, int dim, double * midpoint, double * line, int ihi){
	int i,j;
	for (j=0; j<dim; j++)
		midpoint[j]=0.0;

	for (i=0; i<dim+1; i++)
		if (i!=ihi){
			for (j=0; j<dim; j++)
				midpoint[j]+=simplex[i][j];
		}

	for (j=0; j<dim; j++){
        midpoint[j]/=dim;
		line[j]=simplex[ihi][j] - midpoint[j];
	}
}


// updating the simplex
int Simplex::update_simplex(double * point, int dim, double & fmax, double * midpoint,
				   double * line, double scale, double (*func)(double *, int)){
	int i, update=0;
	double * next=alloc_vector(dim), fx;
	
	for (i=0; i<dim; i++)
		next[i]=midpoint[i]+scale*line[i];
	fx=(*func)(next, dim);

	if(fx<fmax){
		for(i=0; i<dim; i++)
			point[i]=next[i];
		fmax=fx;
		update=1;
	}

	free_vector(next, dim);
	return update;
}

// contracting the simplex
void Simplex::contract_simplex(double ** simplex, int dim, double * fx, 
					  int ilo, double (*func)(double *, int)){
	int i,j;

	for (i=0; i<dim+1; i++)
		if (i!=ilo){
			for (j=0; j<dim; j++)
				simplex[i][j] = (simplex[ilo][j]+simplex[i][j])*0.5;
			fx[i]=(*func)(simplex[i],dim);
		}
}

// checking convergence
int Simplex::check_tol(double fmax, double fmin, double ftol){
	double delta=fabs(fmax-fmin);
	double accuracy = (fabs(fmax)+fabs(fmin))*ftol;

	return (delta<(accuracy+ZEPS));
}

// minimization routine
double Simplex::amoeba(double * point, int coldim, int rowdim, double (*func)(double *, int),double tol){
	
	rowdim = rowdim;
	coldim = coldim;

	int ihi, ilo, inhi, j;
	double fmin;
	double * fx=alloc_vector(coldim+1);
	double * midpoint=alloc_vector(coldim);
	double * line=alloc_vector(coldim);
	double ** simplex=make_simplex(point,coldim);
    
    evaluate_simplex(simplex, coldim, fx, func);

	while (1) {
		simplex_extremes(fx, coldim, ihi, ilo, inhi);
		simplex_bearings(simplex, coldim, midpoint, line, ihi);

		count ++;

		if (check_tol(fx[ihi], fx[ilo], tol)||count>1000) break;

		update_simplex(simplex[ihi], coldim, fx[ihi], midpoint, line, -1.0, func);

		if (fx[ihi]<fx[ilo])
			update_simplex(simplex[ihi], coldim, fx[ihi], midpoint, line, -2.0, func);

		else if (fx[ihi]>=fx[inhi])
			if (!update_simplex(simplex[ihi], coldim, fx[ihi], midpoint, line, 0.5, func))
				contract_simplex(simplex, coldim, fx, ilo, func);
	}

	for (j=0; j< coldim; j++)
		point[j] = simplex[ilo][j];
	fmin=fx[ilo];

	//printf("%lf %d",fmin,count);

	free_vector(fx, coldim);
	free_vector(midpoint, coldim);
	free_vector(line, coldim);
	free_matrix(simplex, coldim+1, coldim);
    
	if (count > 1000) return 0;
	else return fmin;
}




